module Minimax where

import Board
import Heuristics
import System.Random
import Data.List (maximumBy)
import Data.Ord (comparing)
import System.IO.Unsafe (unsafePerformIO)
import Control.Parallel.Strategies

bestMove :: Board -> IO (Maybe Direction)
bestMove board
    | null validMoves = putStrLn "No valid moves available." >> return Nothing
    | otherwise = do
        let best = fst $ maximumBy (comparing snd) validMoves
        putStrLn $ "Best move: " ++ show best
        return $ Just best
  where
    validMoves = filterMoves $ parallelizeMoves [U, D, L, R]
    filterMoves = filter (\(dir, _) -> move dir board /= board)
    parallelizeMoves dirs = map (\dir -> (dir, minimax (move dir board) False 0)) dirs `using` parBuffer 500 rdeepseq

minimax :: Board -> Bool -> Int -> Float
minimax board isMaximizer depth
  | depth == maxDepth || not (canMove board) = evalHeuristic board
  | depth > 4 =
      if isMaximizer
        then maximum [minimax newBoard False (depth + 1) | dir <- [U, D, L, R], let newBoard = move dir board, newBoard /= board]
        else minimax (addRandomTile board) True (depth + 1)
  | otherwise =
      if isMaximizer
        then maximum $ parallelizeMinimax [move dir board | dir <- [U, D, L, R], move dir board /= board] False (depth + 1)
        else minimax (addRandomTile board) True (depth + 1)

parallelizeMinimax boards maximizing depth = map (\b -> minimax b maximizing depth) boards `using` parBuffer 500 rdeepseq

-- Helper function to simulate adding a random tile
addRandomTile :: Board -> Board
addRandomTile board = unsafePerformIO $ do
  let emptyTiles = getEmptyTiles board
  maybePoint <- selectRandomTile emptyTiles
  newValue <- randomTileValue
  return $ case maybePoint of
    Just point -> updateTile point newValue board
    Nothing -> board  -- Return the board as is if no empty tile is available

maxDepth :: Int
maxDepth = 10 -- Adjust the depth as needed

canMove :: Board -> Bool
canMove b = any (\d -> move d b /= b) [U, D, L, R]
